from sarasa import Config, Trainer
from sarasa.optimizers import AdamW, Muon

if __name__ == "__main__":
    # optim:adam-w, optim:muon, ... to select optimizer
    config = Config.from_cli(optim_type=AdamW | Muon)

    trainer = Trainer(config)
    trainer.train()



from collections import defaultdict

import torch
from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import checkpoint_wrapper
from torch.utils.checkpoint import CheckpointPolicy, create_selective_checkpoint_contexts

# for selective op activation checkpointing
_ops_sac_save = {
    torch.ops.aten.mm.default,
    torch.ops.aten._scaled_dot_product_efficient_attention.default,
    torch.ops.aten._scaled_dot_product_flash_attention.default,
    torch.ops.aten._scaled_dot_product_cudnn_attention.default,
    torch.ops.aten._scaled_dot_product_attention_math.default,
    torch.ops.aten._scaled_dot_product_fused_attention_overrideable.default,
    torch.ops._c10d_functional.reduce_scatter_tensor.default,
    # for low precision training, it's useful to always save
    # the result of max, since the absolute maximum is
    # used to compute the scaling factor for quantization.
    torch.ops.aten.max.default,
    torch._higher_order_ops.flex_attention,
    torch.ops.torch_attn._varlen_attn.default,
    torch._higher_order_ops.inductor_compiled_code,
}


def _op_sac_policy(
    ops_to_save: set,
    mm_recompute_shapes: set | None,
    every_nth_mm: int,
):
    mm_recompute_shapes = mm_recompute_shapes or set()

    def _get_custom_policy(meta: dict):
        def _custom_policy(ctx, func, *args, **kwargs):
            # special case, offload to CPU
            if (
                func == torch.ops.aten._to_copy.default
                and "cuda" in str(args[0].device)
                and str(kwargs.get("device", "")) == "cpu"
            ):
                return CheckpointPolicy.MUST_SAVE

            # track mm ops
            mode = "recompute" if ctx.is_recompute else "forward"
            key = f"{mode}_mm_count"

            if func == torch.ops.aten.mm.default:
                if len(args) > 1 and args[1].shape in mm_recompute_shapes:
                    # moe's router
                    return CheckpointPolicy.PREFER_RECOMPUTE
                meta[key] += 1

            # save ops in save list, except every nth mm op
            must_save = (func in ops_to_save) and not (
                func == torch.ops.aten.mm.default and (meta[key] % every_nth_mm == 0)
            )
            return CheckpointPolicy.MUST_SAVE if must_save else CheckpointPolicy.PREFER_RECOMPUTE

        return _custom_policy

    def selective_checkpointing_context_fn():
        return create_selective_checkpoint_contexts(_get_custom_policy(defaultdict(int)))

    return selective_checkpointing_context_fn


def apply_op_sac(
    model: torch.nn.Module,
    ops_to_save: set | None = None,
    mm_recompute_shapes: set | None = None,
    every_nth_mm: int = 2,
) -> torch.nn.Module:
    """Applies selective op activation checkpointing to the given model.

    Ops like mm is expensive, so we want to store their activations for backward.
    On the other hand, ops like activation functions are cheap, so we prefer to recompute them.

    """
    ops_to_save = ops_to_save or _ops_sac_save
    return checkpoint_wrapper(
        model,
        _op_sac_policy(ops_to_save, mm_recompute_shapes, every_nth_mm),  # pyrefly: ignore [bad-argument-type]
    )



from __future__ import annotations

import dataclasses
import enum
from pathlib import Path
from typing import TYPE_CHECKING, Any

from datasets import Dataset as HFDataset
from datasets import IterableDataset as HFIterableDataset
from datasets import load_dataset
from torch.utils.data import DataLoader

from sarasa.data.hf_datasets import HFTextDataset
from sarasa.data.tokenizer import HFTokenizerWrapper

if TYPE_CHECKING:
    from sarasa.config import Evaluate as EvaluateConfig


class Datasets(enum.StrEnum):
    c4 = enum.auto()
    fineweb_edu = enum.auto()
    fineweb_edu_100b = enum.auto()
    fineweb_edu_dedup = enum.auto()

    def load(
        self,
        cache_dir: str | None,
        val_size: int,
    ) -> tuple[HFIterableDataset, HFDataset | None]:
        match self:
            case Datasets.c4:
                ds = load_dataset(
                    "allenai/c4",
                    name="en",
                    split="train",
                    streaming=True,
                    cache_dir=cache_dir,
                )
            case Datasets.fineweb_edu:
                ds = load_dataset(
                    "HuggingFaceFW/fineweb-edu",
                    name="default",
                    split="train",
                    streaming=True,
                    cache_dir=cache_dir,
                )
            case Datasets.fineweb_edu_100b:
                ds = load_dataset(
                    "HuggingFaceFW/fineweb-edu",
                    name="sample-100BT",
                    split="train",
                    streaming=True,
                    cache_dir=cache_dir,
                )
            case Datasets.fineweb_edu_dedup:
                ds = load_dataset(
                    "HuggingFaceTB/smollm-corpus",
                    "fineweb-edu-dedup",
                    split="train",
                    streaming=True,
                    cache_dir=cache_dir,
                )
            case _:
                raise ValueError(f"Unsupported dataset: {self}")

        train_ds, val_ds = ds, None
        if val_size > 0:
            val_ds = HFDataset.from_list(list(ds.take(val_size)))
            train_ds = ds.skip(val_size)
        return train_ds, val_ds


@dataclasses.dataclass
class DataConfig:
    dataset: Datasets = Datasets.fineweb_edu_100b
    """Dataset to use for training. Can be a predefined dataset or a custom dataset path."""

    tokenizer_path: Path | str = Path("./tokenizer")
    """Path to `tokenizer.json` and `tokenizer_config.json` files."""

    seq_len: int = 2048

    num_workers: int = 4

    pin_memory: bool = True

    cache_dir: str | None = None
    """Path to cache directory for datasets. If None, default cache directory is used."""

    def create(
        self,
        batch_size: int,
        val_cfg: EvaluateConfig,
        use_varlen: bool,
    ) -> dict[str, Any]:
        # return {"tokenizer": tokenizer, "train_loader": train_loader, "val_loader": val_loader | None}
        if use_varlen and batch_size != 1:
            raise ValueError("Variable-length batching is only supported with batch_size=1.")
        tokenizer = HFTokenizerWrapper(Path(self.tokenizer_path))
        train_ds, val_ds = self.dataset.load(cache_dir=self.cache_dir, val_size=val_cfg.val_size)
        data_loader = DataLoader(
            HFTextDataset(
                train_ds,
                tokenizer,
                self.seq_len,
                use_varlen=use_varlen,
                strategy="document_pack_pad" if use_varlen else "streaming_pad",
            ),
            batch_size,
            num_workers=self.num_workers,
            pin_memory=self.pin_memory,
        )
        val_loader = None
        if val_ds is not None and val_cfg.val_size > 0:
            val_loader = DataLoader(
                HFTextDataset(
                    val_ds,
                    tokenizer,
                    self.seq_len,
                    infinite=False,
                    use_varlen=use_varlen,
                    strategy="document_pack_pad" if use_varlen else "streaming_pad",
                ),
                batch_size,
                num_workers=self.num_workers,
                pin_memory=self.pin_memory,
            )

        return {
            "tokenizer": tokenizer,
            "train_loader": data_loader,
            "val_loader": val_loader,
        }



from __future__ import annotations

import contextlib
import gc
import os
import sys
import time
import typing
from datetime import timedelta
from functools import cache
from pathlib import Path

import torch
from loguru import logger
from torch import distributed as dist

if typing.TYPE_CHECKING:
    from sarasa.config import Config, Distributed, Profile
    from sarasa.models import BaseModel


IGNORE_INDEX = -100


def setup_profiler(
    config: Profile,
    device: torch.device,
    save_dir: Path,
) -> torch.profiler.profile | contextlib.nullcontext:
    if not config.enabled:

        class DummyProfiler:
            def step(self): ...

        return contextlib.nullcontext(DummyProfiler())

    activities = [torch.profiler.ProfilerActivity.CPU]
    if (pa := getattr(torch.profiler.ProfilerActivity, device.type.upper(), None)) is not None:
        activities.append(pa)

    logger.info(f"Profiler is activated with activities: {[a.name for a in activities]}")

    return torch.profiler.profile(
        activities=activities,
        schedule=torch.profiler.schedule(
            wait=config.wait,
            warmup=config.warmup,
            active=config.active,
            repeat=config.repeat,
        ),
        on_trace_ready=torch.profiler.tensorboard_trace_handler(str(save_dir / "profiler")),
        record_shapes=True,
        profile_memory=True,
        with_stack=True,
    )


def setup_logger(config: Config) -> None:
    logger.remove()
    if config.debug:
        logger_format = f"<blue>RANK={rank()}</blue> | " + (
            "<green>{time:YYYY-MM-DD HH:mm:ss.SSS}</green> | "
            "<level>{level: <8}</level> | "
            "<cyan>{name}</cyan>:<cyan>{function}</cyan>:<cyan>{line}</cyan> | "
            "<level>{message}</level>"
        )
        logger.add(
            sys.stderr,
            format=logger_format,
            backtrace=True,
            diagnose=True,
            level="DEBUG",
        )
    else:
        # log to stderr only for rank 0
        logger.add(sys.stderr, backtrace=True, diagnose=True, level="INFO" if rank() == 0 else "ERROR")

        if world_size() > 1:
            logger.info(
                "Logger is set up for distributed training. Only rank 0 will log messages. Use --debug for more verbose logging."
            )


@contextlib.contextmanager
def set_dtype(
    dtype: torch.dtype,
) -> typing.Generator[None, None, None]:
    old_dtype = torch.get_default_dtype()
    torch.set_default_dtype(dtype)
    try:
        yield
    finally:
        torch.set_default_dtype(old_dtype)


class GarbageCollector:
    def __init__(
        self,
        gc_freq: int,
    ) -> None:
        self.gc_freq = gc_freq
        if self.gc_freq > 0:
            # manually manage gc
            gc.disable()

    def collect(
        self,
        step: int,
    ) -> None:
        if self.gc_freq <= 0:
            # auto gc, nothing to do
            return

        if step % self.gc_freq == 0:
            begin = time.perf_counter()
            gc.collect(generation=1)
            end = time.perf_counter()
            logger.info(f"Garbage collection at step {step} took {end - begin:.4f} seconds")


# distributed utils


@cache
def world_size() -> int:
    if "WORLD_SIZE" in os.environ:
        return int(os.environ["WORLD_SIZE"])

    if not dist.is_initialized():
        return 1
    return dist.get_world_size()


@cache
def rank() -> int:
    if "RANK" in os.environ:
        return int(os.environ["RANK"])

    if not dist.is_initialized():
        return 0
    return dist.get_rank()


def init_distributed(
    backend: str,
    init_timeout_seconds: int,
) -> None:
    if "RANK" in os.environ:
        # run with torchrun
        dist.init_process_group(
            backend=backend,
            timeout=timedelta(seconds=init_timeout_seconds),
        )
        logger.info("Initialized distributed process group")
    else:
        logger.info("Skipping distributed initialization")


def update_timeout(
    timeout_seconds: int,
    device: torch.device,
) -> None:
    logger.info(f"Updating distributed timeout to {timeout_seconds} seconds")
    torch.distributed.barrier(device_ids=[torch.accelerator.current_device_index()])
    torch.accelerator.synchronize(device)

    # at the moment, default process group is the only one supported (None)
    torch.distributed.distributed_c10d._set_pg_timeout(timedelta(seconds=timeout_seconds), None)


def apply_distributed(
    config: Distributed,
    model: BaseModel,
    device: torch.device,
    param_dtype: torch.dtype,
    reduce_dtype: torch.dtype,
    compile: bool,
) -> None:
    mesh = dist.device_mesh.init_device_mesh(device.type, (world_size(),))

    if config.dp_replicate_degree != 1:
        from torch.distributed._composable.replicate import replicate

        if compile:
            torch._dynamo.config.optimize_ddp = "ddp_optimizer"

        # pyrefly: ignore [invalid-param-spec]
        replicate(model, device_mesh=mesh, bucket_cap_mb=100)
        logger.info("Applied DDP to the model")

    elif config.dp_shard_degree == -1:
        from torch.distributed.fsdp import MixedPrecisionPolicy, fully_shard

        mp_policy = MixedPrecisionPolicy(param_dtype=param_dtype, reduce_dtype=reduce_dtype)

        for block in model.blocks:
            fully_shard(block, mesh=mesh, mp_policy=mp_policy, reshard_after_forward=config.reshard_after_forward)
        fully_shard(model, mesh=mesh, mp_policy=mp_policy, reshard_after_forward=config.reshard_after_forward)

        # sarasa scales the loss by valid tokens
        # type: ignore
        model.set_gradient_divide_factor(1.0)

        logger.info(
            f"Applied FSDP to the model (param_dtype={mp_policy.param_dtype}, "
            f"reduce_dtype={mp_policy.reduce_dtype}, reshard_after_forward={config.reshard_after_forward})"
        )



import enum
import gc
import time
from pathlib import Path

import torch
import torch.distributed as dist
import torch.distributed.checkpoint as dcp
from loguru import logger
from torch.distributed.checkpoint.staging import DefaultStager, StagingOptions
from torch.distributed.checkpoint.state_dict import get_model_state_dict, set_model_state_dict
from torch.distributed.checkpoint.state_dict_saver import AsyncCheckpointerType
from torch.distributed.checkpoint.stateful import Stateful

from sarasa.config import Config


class AsyncMode(enum.StrEnum):
    none = enum.auto()
    default = enum.auto()
    mem_pinned = enum.auto()


class ModelWrapper(Stateful):
    def __init__(self, model: torch.nn.Module):
        self.model = model

    def state_dict(self) -> dict[str, dict]:
        return {"model": get_model_state_dict(self.model)}

    def load_state_dict(self, state_dict: dict[str, dict]) -> None:
        set_model_state_dict(self.model, state_dict["model"])


class Checkpointer:
    def __init__(
        self,
        config: Config,
        model: torch.nn.Module,
    ):
        self.config = config
        self.freq = config.checkpoint.freq
        self.checkpoint_dir = Path(config.output_dir) / "checkpoints"
        self.checkpoint_dir.mkdir(parents=True, exist_ok=True)
        self.async_mode = AsyncMode(config.checkpoint.async_mode)
        if self.async_mode != AsyncMode.none:
            self.pg = dist.new_group(backend="gloo") if dist.is_initialized() else None

        self.stager = None
        self.save_future = None
        self.stage_future = None

        self.state = ModelWrapper(model)

    def trigger(
        self,
        step: int,
    ) -> bool:
        return step % self.freq == 0

    @torch.no_grad()
    def load(
        self,
        step: int,
    ) -> None:
        checkpoint_id = str(self.checkpoint_dir / f"checkpoint_{step:09d}")
        state_dict = self.state.state_dict()
        dcp.load(
            state_dict=state_dict,
            checkpoint_id=checkpoint_id,
        )

    @torch.no_grad()
    def save(
        self,
        step: int,
    ) -> None:
        begin = time.perf_counter()
        checkpoint_id = str(self.checkpoint_dir / f"checkpoint_{step:09d}")

        # todo: save other states
        state_dict = self.state.state_dict()

        if self.async_mode == AsyncMode.default:
            gc.collect(1)
            if self.save_future is not None:
                self.save_future.result()
            self.save_future = dcp.async_save(
                state_dict,
                storage_writer=None,
                checkpoint_id=checkpoint_id,
                process_group=self.pg,  # pyrefly: ignore [bad-argument-type]
            )
            gc.collect(1)
        elif self.async_mode == AsyncMode.mem_pinned:
            gc.collect(1)
            if self.save_future is not None:
                self.save_future.result()
            if self.stager is None:
                self.stager = DefaultStager(StagingOptions(True, True, True, True))
            ret = dcp.async_save(
                state_dict,
                storage_writer=None,
                checkpoint_id=checkpoint_id,
                process_group=self.pg,  # pyrefly: ignore [bad-argument-type]
                async_checkpointer_type=AsyncCheckpointerType.PROCESS,
                async_stager=self.stager,
            )
            self.save_future = ret.upload_completion  # pyrefly: ignore
            self.stage_future = ret.staging_completion  # pyrefly: ignore
        else:
            ret = dcp.save(
                state_dict,
                storage_writer=None,
                checkpoint_id=checkpoint_id,
            )

        logger.info(f"Finished saving checkpoint at step {step} in {time.perf_counter() - begin:.2f} seconds")

    def wait_for_staging(self) -> None:
        # no-op if not using mem_pinned async mode
        if self.stage_future is not None:
            self.stage_future.result()

    def close(self) -> None:
        if self.stager is not None:
            self.stager.close()

        if self.save_future is not None:
            self.save_future.result()

        if self.pg is not None:
            dist.destroy_process_group(self.pg)  # pyrefly: ignore [bad-argument-type]
            self.pg = None

    def __del__(self):
        self.close()



import dataclasses
from typing import Literal, cast

import torch

from sarasa.models import BaseModel
from sarasa.optimizers.utils import GroupedOptimizer


@dataclasses.dataclass
class AdamW:
    """
    Default optimizer
    """

    lr: float = 1e-4
    weight_decay: float = 0.1
    betas: tuple[float, float] = (0.9, 0.95)

    def create(
        self,
        model: BaseModel,
    ) -> torch.optim.Optimizer:
        param_groups = model.param_groups()
        params = sum(param_groups.values(), [])
        optimizer = torch.optim.AdamW(
            params,
            lr=torch.tensor(self.lr, dtype=torch.float32),
            weight_decay=self.weight_decay,
            betas=self.betas,
            fused=True,
        )
        return optimizer


@dataclasses.dataclass
class Muon:
    """
    Muon optimizer
    """

    lr: float = 1e-4
    weight_decay: float = 0.1
    momentum: float = 0.9

    adam_lr: float | None = None
    adam_betas: tuple[float, float] = (0.9, 0.95)
    adam_weight_decay: float = 0

    adjust_lr_fn: Literal["original", "match_rms_adamw"] = "match_rms_adamw"

    def __post_init__(self):
        self.adam_lr = self.adam_lr or self.lr

    def create(
        self,
        model: BaseModel,
    ) -> torch.optim.Optimizer:
        param_groups = model.param_groups()

        muon = torch.optim.Muon(
            param_groups["matrix"],
            lr=self.lr,
            weight_decay=self.weight_decay,
            momentum=self.momentum,
            adjust_lr_fn=self.adjust_lr_fn,
        )

        adam = torch.optim.AdamW(
            sum([param_groups[k] for k in param_groups if k != "matrix"], []),
            lr=cast(float, self.adam_lr),
            betas=self.adam_betas,
            weight_decay=self.adam_weight_decay,
            fused=True,
        )

        return GroupedOptimizer(muon, adam)



import torch


class GroupedOptimizer(torch.optim.Optimizer):
    def __init__(
        self,
        *optimizers: torch.optim.Optimizer,
    ):
        super().__init__(sum([optim.param_groups for optim in optimizers], []), {})
        self.optimizers = optimizers

    # pyrefly: ignore [bad-override]
    def step(
        self,
    ) -> None:
        for optim in self.optimizers:
            optim.step()

    def zero_grad(
        self,
        set_to_none: bool = True,
    ) -> None:
        for optim in self.optimizers:
            optim.zero_grad(set_to_none=set_to_none)

    def state_dict(self) -> dict:
        return super().state_dict()

    def load_state_dict(self, state_dict: dict) -> None:
        super().load_state_dict(state_dict)



import dataclasses
import sys

import pytest

from sarasa import Config


def test_config_custom_model_type(monkeypatch):
    monkeypatch.setattr(sys, "argv", ["program"])  # avoid checking actual sys.argv

    @dataclasses.dataclass
    class CustomModelConfig:
        param: int = 42

    cfg = Config.from_cli(model_type=CustomModelConfig)
    assert cfg.model.param == 42


def test_config_complex_custom_model_type(monkeypatch):
    monkeypatch.setattr(sys, "argv", ["program", "model:llama3", "--model.param", "100"])

    @dataclasses.dataclass
    class Llama3:
        param: int = 42

    @dataclasses.dataclass
    class Qwen3:
        param: str = "hello"

    cfg = Config.from_cli(model_type=Llama3 | Qwen3)
    assert isinstance(cfg.model, Llama3)
    assert cfg.model.param == 100


@pytest.fixture
def config_py(tmp_path, num_configs) -> str:
    file = tmp_path / "config.py"
    lines = ["from sarasa.config import *"]
    for i in range(num_configs):
        lines.append(f"config{i} = Config(Model(), AdamW(), LRScheduler(), Data(), checkpoint=Checkpoint(freq=10))")
    with open(file, "w") as f:
        f.write("\n".join(lines))
    yield str(file)


@pytest.mark.parametrize("num_configs", [1])
def test_config_loading(config_py, monkeypatch):
    monkeypatch.setattr(sys, "argv", ["program", "--config_file", config_py])
    cfg = Config.from_cli()
    assert cfg.checkpoint.freq == 10


@pytest.mark.parametrize("num_configs", [1])
def test_config_loading_overriding(config_py, monkeypatch):
    monkeypatch.setattr(sys, "argv", ["program", "--config_file", config_py, "--checkpoint.freq", "100"])
    cfg = Config.from_cli()
    assert cfg.checkpoint.freq == 100


@pytest.mark.parametrize("num_configs", [0, 2])
def test_config_loading_content_error(config_py, monkeypatch):
    monkeypatch.setattr(sys, "argv", ["program", "--config_file", config_py])
    with pytest.raises(ValueError):
        Config.from_cli()


def test_config_loading_filetype_error(monkeypatch, tmp_path):
    config_file = tmp_path / "config.json"
    config_file.write_text("{}")
    monkeypatch.setattr(sys, "argv", ["program", "--config_file", str(config_file)])
    with pytest.raises(ValueError):
        Config.from_cli()



from __future__ import annotations

import dataclasses
import sys
from pathlib import Path
from typing import Literal

import torch

from sarasa.utils import world_size

"""
Variable configuration dataclasses for model, optimizer, lr scheduler, and data
These classes have `create` methods to instantiate the actual objects

Users can define their own configuration dataclasses and pass them to Config.from_cli to use custom components
"""
from sarasa.data import DataConfig as Data  # noqa
from sarasa.models import ModelConfig as Model  # noqa
from sarasa.optimizers import AdamW  # noqa


@dataclasses.dataclass
class _WithSeqLen:
    seq_len: int | None = None


@dataclasses.dataclass
class LRScheduler:
    warmup_steps: int = 200
    decay_ratio: float | None = None
    """If set, the ratio of total steps to apply decay after warmup. If None, decay starts immediately after warmup."""

    decay_type: Literal["linear", "cosine", "sqrt"] = "linear"
    min_lr_factor: float = 0.0

    def create(
        self,
        optimizer: torch.optim.Optimizer,
        total_iters: int,
    ) -> torch.optim.lr_scheduler.SequentialLR:
        assert self.decay_ratio is None or (0 <= self.decay_ratio <= 1), "decay_ratio must be between 0 and 1"
        warmup_steps = self.warmup_steps
        stay_steps = 0 if self.decay_ratio is None else int(total_iters * (1 - self.decay_ratio)) - warmup_steps
        decay_steps = total_iters - warmup_steps - stay_steps
        assert warmup_steps >= 0 and decay_steps >= 0 and stay_steps >= 0, (
            f"Invalid lr scheduler steps configuration: {warmup_steps=}, {decay_steps=}, {stay_steps=}"
        )

        # 1 / max(1, warmup_steps) to avoid division by zero
        warmup = torch.optim.lr_scheduler.LinearLR(optimizer, 1 / max(1, warmup_steps), total_iters=warmup_steps)

        stay = torch.optim.lr_scheduler.ConstantLR(optimizer=optimizer, factor=1.0, total_iters=stay_steps)

        match self.decay_type:
            case "linear":
                decay = torch.optim.lr_scheduler.LinearLR(
                    optimizer,
                    start_factor=1.0,
                    end_factor=self.min_lr_factor,
                    total_iters=decay_steps,
                )
            case "sqrt":
                decay = torch.optim.lr_scheduler.LambdaLR(
                    optimizer,
                    lr_lambda=lambda step: (
                        max(
                            self.min_lr_factor,
                            (decay_steps - step) / decay_steps,
                        )
                        ** 0.5
                    ),
                )
            case "cosine":
                decay = torch.optim.lr_scheduler.CosineAnnealingLR(
                    optimizer,
                    T_max=decay_steps,
                    eta_min=optimizer.param_groups[0]["lr"] * self.min_lr_factor,
                )
            case _:
                raise ValueError(f"Unsupported decay type: {self.decay_type}")

        scheduler = torch.optim.lr_scheduler.SequentialLR(
            optimizer,
            [warmup, stay, decay],
            [self.warmup_steps, self.warmup_steps + stay_steps],
        )
        return scheduler


"""
Static configuration dataclasses

These classes are not expected to be changed by the user
"""


@dataclasses.dataclass
class Train:
    steps: int = 10_000

    grad_clip: float | None = None

    dtype: Literal["bfloat16", "float32"] = "float32"
    """Dtype used for model initialization"""

    amp_dtype: Literal["bfloat16", "float16", "float32"] = "bfloat16"
    """Dtype used for automatic mixed precision training"""

    compile: bool = False

    gc_freq: int = 500
    """Garbage collection frequency (in steps). If -1, no periodic GC is performed."""

    local_batch_size: int = 32
    """local (per device) batch size"""

    global_batch_size: int = 256
    """
    global (across all devices) batch size, used to compute 
    grad_accum_steps = global_batch_size // (local_batch_size * num_devices)
    """

    use_fa4: bool = False
    """Whether to use FA4 flash attention if available."""

    use_sac: bool = False
    """Whether to use selective activation checkpointing."""

    def __post_init__(self):
        assert self.global_batch_size % self.local_batch_size == 0
        assert self.global_batch_size % world_size() == 0
        assert self.global_batch_size >= self.local_batch_size * world_size()


@dataclasses.dataclass
class Evaluate:
    freq: int = -1
    """Evaluation frequency (in steps). If -1, no evaluation is performed."""

    val_size: int = 0
    """Number of samples in the evaluation set. If 0, no evaluation is performed.
    Must be divisible by (local_batch_size * num_devices)."""

    local_batch_size: int | None = None
    """local (per device) batch size for evaluation. If None, uses train.local_batch"""


@dataclasses.dataclass
class Profile:
    enabled: bool = False
    wait: int = 10  # the number of steps to wait before warming up the profiler
    warmup: int = 10  # the number of steps to warm up the profiler before starting to record
    active: int = 10  # the number of steps for the active record
    repeat: int = 1  # the number of cycles, 0 means always


@dataclasses.dataclass
class Metrics:
    freq: int = 10
    use_tensorboard: bool = False
    all_node: bool = False


@dataclasses.dataclass
class Checkpoint:
    freq: int = 1000
    async_mode: Literal["none", "default", "mem_pinned"] = "default"


@dataclasses.dataclass
class Distributed:
    dp_replicate_degree: int = 1
    """Degree of DDP. 1 means no replication"""

    dp_shard_degree: int = -1
    """Degree of FSDP. -1 means full sharding"""

    backend: Literal["nccl", "gloo"] = "nccl"

    init_timeout_seconds: int = 300
    """Timeout for initializing the distributed process group."""

    train_timeout_seconds: int = 100
    """Timeout for distributed training operations after the first iteration."""

    reshard_after_forward: bool = False
    """Whether to reshard model parameters after each forward pass (FSDP only)."""

    def __post_init__(self):
        if not (self.dp_replicate_degree == 1 and self.dp_shard_degree == -1):
            raise NotImplementedError()


@dataclasses.dataclass
class Config[ModelT: _WithSeqLen, OptimizerT, LRSchedulerT, DataT: _WithSeqLen]:
    # variable components
    model: ModelT
    optim: OptimizerT
    lr_scheduler: LRSchedulerT
    data: DataT

    # static components
    train: Train = dataclasses.field(default_factory=Train)
    evaluate: Evaluate = dataclasses.field(default_factory=Evaluate)
    profile: Profile = dataclasses.field(default_factory=Profile)
    metrics: Metrics = dataclasses.field(default_factory=Metrics)
    checkpoint: Checkpoint = dataclasses.field(default_factory=Checkpoint)
    distributed: Distributed = dataclasses.field(default_factory=Distributed)

    seed: int = 0
    debug: bool = False
    """ Enable debug mode with more verbose logging and checks."""

    output_dir: Path = Path("./outputs")
    """Directory to save checkpoints and logs."""

    config_file: Path | None = None
    """Path to a config file (JSON or TOML) to load configuration from."""

    def __post_init__(self):
        if self.output_dir is not None:
            self.output_dir.mkdir(parents=True, exist_ok=True)

        if hasattr(self.model, "seq_len"):
            if self.model.seq_len is None and self.data.seq_len is not None:
                self.model.seq_len = self.data.seq_len
            if self.model.seq_len is None:
                raise ValueError("seq_len must be specified in either model or data configuration.")

        if self.evaluate.local_batch_size is None:
            self.evaluate.local_batch_size = self.train.local_batch_size

        if self.evaluate.freq > 0:
            assert self.evaluate.val_size > 0, "evaluate.val_size must be > 0 if evaluate.freq > 0"
            assert self.evaluate.val_size % (self.evaluate.local_batch_size * world_size()) == 0, (
                "evaluate.val_size must be divisible by (evaluate.local_batch_size * num_devices)"
            )

    @classmethod
    def create(
        cls,
        model: ModelT,
        optim: OptimizerT,
        lr_scheduler: LRSchedulerT,
        data: DataT,
        **kwargs,
    ) -> Config:
        return cls(
            model=model,
            optim=optim,
            lr_scheduler=lr_scheduler,
            data=data,
            **kwargs,
        )

    @classmethod
    def from_cli(
        cls,
        *,
        model_type: ModelT = Model,  # pyrefly: ignore [bad-function-definition]
        optim_type: OptimizerT = AdamW,  # pyrefly: ignore [bad-function-definition]
        lr_scheduler_type: LRSchedulerT = LRScheduler,  # pyrefly: ignore [bad-function-definition]
        data_type: DataT = Data,  # pyrefly: ignore [bad-function-definition]
    ) -> Config:
        """
        initialize JobConfig from command line arguments
        update the values with the following priority: CLI arguments > config file > defaults

        *_type can be used to specify custom dataclass types for each section
        >> config = Config.from_cli(optim_type=CustomOptimizerConfig)
        """

        import importlib.util

        import tyro

        from sarasa.utils import rank

        loaded_config = None

        if (under := ("--config_file" in sys.argv)) or ("--config-file" in sys.argv):
            config_file = sys.argv[sys.argv.index("--config_file" if under else "--config-file") + 1]
            config_file = Path(config_file)

            if not config_file.exists():
                raise FileNotFoundError(f"Config file {config_file} does not exist.")

            if config_file.suffix != ".py":
                raise ValueError("Only Python config files are supported in this method.")

            spec = importlib.util.spec_from_file_location("custom_config", config_file)
            if spec is not None and spec.loader is not None:
                module = importlib.util.module_from_spec(spec)
                spec.loader.exec_module(module)
            configs = [
                config
                for config in module.__dict__.values()
                if isinstance(config, cls) and not isinstance(config, type)
            ]
            if len(configs) == 0:
                raise ValueError(f"No Config instance found in {config_file}.")
            elif len(configs) > 1:
                raise ValueError(f"Multiple Config instances found in {config_file}. Please keep only one.")
            else:
                loaded_config = configs[0]

        return tyro.cli(
            # pyrefly: ignore [unsupported-operation]
            cls[
                model_type,
                optim_type,
                lr_scheduler_type,
                data_type,
            ],
            default=loaded_config,
            console_outputs=(rank() == 0),
            config=(tyro.conf.UsePythonSyntaxForLiteralCollections,),
            # UsePythonSyntaxForLiteralCollections allows dict/list/tuple to be specified in CLI arguments
            # e.g., --model.extra '{"key": "value"}' or --model.extra '["val1", "val2"]'
        )


__all__ = [
    "Config",
    "Model",
    "AdamW",
    "LRScheduler",
    "Data",
    "Train",
    "Evaluate",
    "Metrics",
    "Checkpoint",
    "Distributed",
]



import typing

import torch
from torch import nn
from torch.nn import functional as F

from sarasa.models import BaseModel, ModelConfig
from sarasa.models.attention import CausalSelfAttention, VarlenMetaData
from sarasa.models.utils import RoPE


class MLP(nn.Module):
    def __init__(
        self,
        config: ModelConfig,
        multiple_of: int,
        ffn_dim_multiplier: float | None,
    ):
        super().__init__()
        config.hidden_dim = typing.cast(int, config.hidden_dim)
        hidden_dim = int(8 * config.hidden_dim / 3)
        # custom dim factor multiplier
        if ffn_dim_multiplier is not None:
            hidden_dim = int(ffn_dim_multiplier * hidden_dim)
        hidden_dim = multiple_of * ((hidden_dim + multiple_of - 1) // multiple_of)

        self.w1 = nn.Linear(config.hidden_dim, hidden_dim, bias=False)
        self.w2 = nn.Linear(hidden_dim, config.hidden_dim, bias=False)
        self.w3 = nn.Linear(config.hidden_dim, hidden_dim, bias=False)

    def forward(self, x):
        return self.w2(F.silu(self.w1(x)) * self.w3(x))


class Block(nn.Module):
    def __init__(
        self,
        config: ModelConfig,
        layer_idx: int,
        multiple_of: int,
        ffn_dim_multiplier: float | None,
    ):
        super().__init__()
        self.layer_idx = layer_idx
        self.attention = CausalSelfAttention(config)
        self.mlp = MLP(config, multiple_of, ffn_dim_multiplier)
        config.hidden_dim = typing.cast(int, config.hidden_dim)
        self.attn_norm = nn.RMSNorm(config.hidden_dim, eps=config.rms_eps)
        self.mlp_norm = nn.RMSNorm(config.hidden_dim, eps=config.rms_eps)

    def forward(
        self,
        x: torch.Tensor,
        cos_sin: tuple[torch.Tensor, torch.Tensor],
        *,
        metadata: VarlenMetaData | None = None,
    ) -> torch.Tensor:
        x = x + self.attention(self.attn_norm(x), cos_sin, metadata=metadata)
        x = x + self.mlp(self.mlp_norm(x))
        return x


class Llama3(BaseModel):
    def __init__(
        self,
        config: ModelConfig,
    ):
        super().__init__(config)
        multiple_of = int(config.extra.get("multiple_of", 1024))
        ffn_dim_multiplier = float(config.extra.get("ffn_dim_multiplier", 1.4))

        self.token_emb = nn.Embedding(config.vocab_size, config.hidden_dim)  # type: ignore
        self.max_seq_len = config.seq_len * 16  # type: ignore
        self.head_dim = config.head_dim
        cos, sin = RoPE.precompute(self.max_seq_len, config.head_dim)
        self.register_buffer("cos", cos, persistent=False)
        self.register_buffer("sin", sin, persistent=False)
        self.blocks = nn.ModuleList([
            Block(config, layer_idx, multiple_of, ffn_dim_multiplier) for layer_idx in range(config.num_layers)
        ])
        self.norm = nn.RMSNorm(config.hidden_dim, eps=config.rms_eps)  # type: ignore
        self.output = nn.Linear(config.hidden_dim, config.vocab_size, bias=False)  # type: ignore

    @torch.no_grad()
    def init_weights(self) -> None:
        self.cos, self.sin = RoPE.precompute(self.max_seq_len, self.head_dim, device=self.cos.device)
        torch.nn.init.normal_(self.token_emb.weight)
        # pyrefly: ignore [bad-assignment]
        for block in self.blocks:
            block: Block
            init_std = 0.02 / (2 * (block.layer_idx + 1)) ** 0.5

            nn.init.trunc_normal_(block.attention.c_q.weight, std=0.02)
            nn.init.trunc_normal_(block.attention.c_k.weight, std=0.02)
            nn.init.trunc_normal_(block.attention.c_v.weight, std=0.02)
            nn.init.trunc_normal_(block.attention.c_proj.weight, std=init_std)

            nn.init.trunc_normal_(block.mlp.w1.weight, std=0.02)
            nn.init.trunc_normal_(block.mlp.w2.weight, std=init_std)
            nn.init.trunc_normal_(block.mlp.w3.weight, std=init_std)

        final_out_std = self.output.weight.shape[-1] ** -0.5
        cutoff_factor = 3
        nn.init.trunc_normal_(
            self.output.weight,
            mean=0.0,
            std=final_out_std,
            a=-cutoff_factor * final_out_std,
            b=cutoff_factor * final_out_std,
        )

        for mod in self.modules():
            if isinstance(mod, nn.RMSNorm):
                mod.reset_parameters()

    def param_groups(self) -> dict[str, list[nn.Parameter]]:
        matrix_params = [param for param in self.blocks.parameters() if param.ndim == 2]
        embedding_params = list(self.token_emb.parameters())
        lm_head_params = list(self.output.parameters())
        rms_norm_params = [
            mod.weight for mod in self.modules() if isinstance(mod, nn.RMSNorm) and mod.elementwise_affine
        ]
        assert len(list(self.parameters())) == (
            len(matrix_params) + len(embedding_params) + len(lm_head_params) + len(rms_norm_params)
        )

        return {
            "matrix": matrix_params,
            "embedding": embedding_params,
            "lm_head": lm_head_params,
            "rms_norm": rms_norm_params,
        }

    def forward(
        self,
        input: torch.Tensor,
        *,
        metadata: VarlenMetaData | None = None,
    ) -> torch.Tensor:
        B, T = input.size()
        x = self.token_emb(input)  # (B, T, C)
        cos_sin = self.cos[:, :T], self.sin[:, :T]

        for block in self.blocks:
            x = block(x, cos_sin, metadata=metadata)

        x = self.norm(x)
        logits = self.output(x)
        return logits



from sarasa.config import AdamW, Config, Data, Evaluate, LRScheduler, Model, Train

config = Config.create(
    model=Model(
        name="nanochat_gpt",
        num_layers=12,
        qk_norm=True,
    ),
    data=Data(tokenizer_path="./tokenizer"),
    lr_scheduler=LRScheduler(
        decay_type="linear",
        warmup_steps=0,
    ),
    optim=AdamW(lr=3e-4),
    train=Train(
        local_batch_size=16,
        global_batch_size=256,
    ),
    evaluate=Evaluate(
        freq=1000,
        val_size=8192,
    ),
    seed=0,
)



import torch

from sarasa.utils import set_dtype


def test_set_dtype() -> None:
    torch.set_default_dtype(torch.float64)
    a = torch.randn(1)
    assert a.dtype == torch.float64

    with set_dtype(torch.float32):
        a = torch.randn(1)
        assert a.dtype == torch.float32

    a = torch.randn(1)
    assert a.dtype == torch.float64



from __future__ import annotations

import abc
import dataclasses
from typing import TYPE_CHECKING, Literal

import torch
from loguru import logger
from torch import nn

if TYPE_CHECKING:
    from sarasa.models.attention import VarlenMetaData


@dataclasses.dataclass
class ModelConfig:
    name: Literal["nanochat_gpt", "llama3"] = "nanochat_gpt"
    """ Name of the model architecture to use.

    - nanochat_gpt: GPT architecture in NanoChat
    - llama3: LLaMA 3 architecture ()
    """

    num_layers: int = 12
    head_dim: int = 128

    num_heads: int | None = None  # inferred later if None
    num_kv_heads: int | None = None  # inferred later if None
    hidden_dim: int | None = None  # inferred later if None
    vocab_size: int | None = None  # set later based on tokenizer
    seq_len: int | None = None  # set later based on data config
    qk_norm: bool = False  # whether to use RMSNorm on q/k
    rms_eps: float | None = None  # epsilon for RMSNorm, default to library default if None
    rms_learnable: bool = False  # whether RMSNorm has learnable scale parameter

    attn_type: Literal["sdpa", "varlen"] = "sdpa"
    """ Attention type, either standard dense attention (sdpa) or variable-length attention (varlen)"""

    extra: dict[str, int | float | bool | str] = dataclasses.field(default_factory=dict)
    """ Extra model-specific configurations. 
    Expected to be used in config files, but can be updated in CLI,
    like --model.extra '{"multiple_of": 1024, "ffn_dim_multiplier": 1.4}'.
    """

    def __post_init__(self):
        # infer hidden_dim, num_heads, num_kv_heads if not provided using the rules presented in nanochat
        self.hidden_dim = self.hidden_dim or (self.num_layers * 64 + self.head_dim - 1) // self.head_dim * self.head_dim
        self.num_heads = self.num_heads or self.hidden_dim // self.head_dim
        self.num_kv_heads = self.num_kv_heads or self.num_heads

        # sanity checks
        assert self.hidden_dim % self.head_dim == 0
        assert self.head_dim * self.num_heads == self.hidden_dim
        assert self.num_kv_heads <= self.num_heads and self.num_heads % self.num_kv_heads == 0

    def create(self) -> BaseModel:
        if self.vocab_size is None or self.seq_len is None:
            raise ValueError("vocab_size and seq_len must be set before creating the model")

        match self.name:
            case "nanochat_gpt":
                from .nanochat_gpt import GPT

                if not self.qk_norm:
                    logger.warning("nanochat_gpt model without qk_norm is not recommended")

                return GPT(self)

            case "llama3":
                from .llama3 import Llama3

                if self.qk_norm:
                    logger.warning("llama3 model with qk_norm is not standard")

                return Llama3(self)

            case _:
                raise ValueError(f"Unknown model name: {self.name}")


class BaseModel(nn.Module, abc.ABC):
    # Common interface for all models in Sarasa

    token_emb: nn.Embedding  # token embedding layer
    blocks: nn.ModuleList  # TF blocks
    config: ModelConfig

    def __init__(
        self,
        config: ModelConfig,
    ):
        super().__init__()
        self.config = config

    @abc.abstractmethod
    @torch.no_grad()
    def init_weights(self) -> None:
        # Actual initialization of model weights
        pass

    @abc.abstractmethod
    def param_groups(self) -> dict[str, list[nn.Parameter]]:
        # Return parameter groups for optimizer
        pass

    @property
    def num_params_flops(
        self,
    ) -> tuple[int, int]:
        # Return number of parameters and FLOPs per token (for dense model)
        config = self.config

        # for tied embeddings, num_params -= num_params_emb
        num_params = sum(p.numel() for p in self.parameters())
        num_params_emb = self.token_emb.weight.numel()

        # If forward pass has 1 matmul, then backward pass has 2 matmuls
        # Each self-attention has 2 matmuls
        num_flops_per_token = 6 * (
            (num_params - num_params_emb) + (config.num_layers * config.num_heads * config.head_dim * config.seq_len)  # type: ignore
        )

        return num_params, num_flops_per_token

    @abc.abstractmethod
    def forward(
        self,
        input: torch.Tensor,
        *,
        metadata: VarlenMetaData | None = None,
    ) -> torch.Tensor:
        # Forward pass of the model, input is (B, T) token ids, output is (B, T, vocab_size) logits
        # metadata is used for, e.g., variable-length attention
        pass



from contextlib import AbstractContextManager
from typing import Callable, Iterator

import torch
import torch.distributed as dist

from sarasa.config import Evaluate as EvaluateConfig
from sarasa.metrics import MetricsProcessor
from sarasa.utils import IGNORE_INDEX, world_size


class Evaluator:
    def __init__(
        self,
        config: EvaluateConfig,
        val_loader: Iterator[tuple[dict[str, torch.Tensor], torch.Tensor]],
        amp_context: AbstractContextManager,
        metrics_processor: MetricsProcessor,
        loss_fn: Callable[[torch.Tensor, torch.Tensor], torch.Tensor],
        device: torch.device,
    ):
        self.freq = config.freq
        self.val_loader = val_loader
        self.amp_context = amp_context
        self.metrics_processor = metrics_processor
        self.loss_fn = loss_fn
        self.device = device

    def trigger(
        self,
        step: int,
    ) -> bool:
        return step % self.freq == 0

    @torch.no_grad()
    def evaluate(
        self,
        model: torch.nn.Module,
        step: int,
    ) -> None:
        model.eval()

        losses = []
        total_tokens = torch.zeros(1, device=self.device)

        for input_dict, target in self.val_loader:
            self.metrics_processor.ntokens_since_last_log += target.numel()
            input_dict = {
                k: v.to(self.device, non_blocking=(self.device.type == "cuda")) for k, v in input_dict.items()
            }
            target = target.to(self.device, non_blocking=(self.device.type == "cuda"))
            valid_tokens = (target != IGNORE_INDEX).sum()

            with self.amp_context:
                pred = model(**input_dict)
                loss = self.loss_fn(pred, target)

            if world_size() > 1:
                dist.all_reduce(valid_tokens, op=dist.ReduceOp.SUM)
                dist.all_reduce(loss, op=dist.ReduceOp.SUM)

            losses.append(loss)
            total_tokens += valid_tokens

        loss = torch.stack(losses).sum() / total_tokens

        self.metrics_processor.val_log(step=step, val_loss=loss.item())

        model.train()



import torch

from sarasa.data.utils import prepare_varlen_metadata


def test_create_varlen_metadata():
    input = torch.tensor([0, 1, 1, 1, 0, 1, 1], dtype=torch.long)
    input_dict = {"input": input}
    bos_token_id = 0
    input_dict = prepare_varlen_metadata(input_dict, bos_token_id)
    metadata = input_dict["metadata"]
    assert metadata.cu_seq_q.tolist() == [0, 4, 7]
    assert metadata.cu_seq_k.tolist() == [0, 4, 7]
    assert metadata.max_q == 4
    assert metadata.max_k == 4



from sarasa.config import AdamW, Config, Data, Evaluate, LRScheduler, Model, Train

config = Config.create(
    model=Model(
        name="llama3",
        hidden_dim=2048,
        num_layers=16,
        num_heads=32,
        num_kv_heads=8,
        head_dim=64,
        rms_eps=1e-5,
        rms_learnable=True,
    ),
    data=Data(tokenizer_path="./tokenizer"),
    lr_scheduler=LRScheduler(
        decay_type="linear",
        warmup_steps=0,
    ),
    optim=AdamW(lr=3e-4),
    train=Train(
        local_batch_size=32,
        global_batch_size=1024,
        use_sac=True,
    ),
    evaluate=Evaluate(
        freq=1000,
        val_size=8192,
    ),
    seed=0,
)



# NanoChat's GPT model, adapted from https://github.com/karpathy/nanochat


import typing

import torch
from loguru import logger
from torch import nn
from torch.nn import functional as F

from sarasa.models import BaseModel, ModelConfig
from sarasa.models.attention import CausalSelfAttention, VarlenMetaData
from sarasa.models.utils import RoPE


class RMSNorm(torch.nn.RMSNorm):
    # RMSNorm without affine parameters
    def __init__(
        self,
        normalized_shape: int,
    ):
        super().__init__(normalized_shape, eps=None, elementwise_affine=False)


class MLP(nn.Module):
    def __init__(
        self,
        config: ModelConfig,
    ):
        super().__init__()
        config.hidden_dim = typing.cast(int, config.hidden_dim)
        self.c_fc = nn.Linear(config.hidden_dim, 4 * config.hidden_dim, bias=False)
        self.c_proj = nn.Linear(4 * config.hidden_dim, config.hidden_dim, bias=False)

    def forward(self, x):
        x = self.c_fc(x)
        x = F.relu(x).square()
        x = self.c_proj(x)
        return x


class Block(nn.Module):
    def __init__(
        self,
        config: ModelConfig,
        layer_idx: int,
    ):
        super().__init__()
        self.attn = CausalSelfAttention(config, layer_idx)
        self.mlp = MLP(config)
        config.hidden_dim = typing.cast(int, config.hidden_dim)
        self.norm = RMSNorm(config.hidden_dim)

    def forward(
        self,
        x: torch.Tensor,
        cos_sin: tuple[torch.Tensor, torch.Tensor],
        *,
        metadata: VarlenMetaData | None = None,
    ) -> torch.Tensor:
        x = x + self.attn(self.norm(x), cos_sin, metadata=metadata)
        x = x + self.mlp(self.norm(x))
        return x


class GPT(BaseModel):
    def __init__(
        self,
        config: ModelConfig,
    ):
        super().__init__(config)
        pad_vocab_size_to = int(config.extra.get("pad_vocab_size_to", 64))
        self.use_softcap = bool(config.extra.get("use_softcap", False))
        use_resid_lambdas = bool(config.extra.get("use_resid_lambdas", False))
        use_x0_lambdas = bool(config.extra.get("use_x0_lambdas", False))

        self.num_heads = typing.cast(int, config.num_heads)
        self.hidden_dim = typing.cast(int, config.hidden_dim)
        self.seq_len = typing.cast(int, config.seq_len)
        self.vocab_size = typing.cast(int, config.vocab_size)
        self.num_layers = config.num_layers

        # For DDP, we want vocab_size divisible by world_size. Also, there are potential performance benefits, see:
        # https://huggingface.co/docs/transformers/main_classes/model#transformers.PreTrainedModel.resize_token_embeddings
        padded_vocab_size = ((self.vocab_size + pad_vocab_size_to - 1) // pad_vocab_size_to) * pad_vocab_size_to
        if padded_vocab_size != self.vocab_size:
            logger.warning(
                f"Padding vocab_size from {self.vocab_size} to {padded_vocab_size} to be divisible by {pad_vocab_size_to}"
            )
        self.token_emb = nn.Embedding(padded_vocab_size, self.hidden_dim)
        self.blocks = nn.ModuleList([Block(config, layer_idx) for layer_idx in range(self.num_layers)])
        self.lm_head = nn.Linear(self.hidden_dim, padded_vocab_size, bias=False)
        self.norm = RMSNorm(self.hidden_dim)

        # Per-layer learnable scalars (inspired by modded-nanogpt)
        # resid_lambdas: scales the residual stream at each layer (init 1.0 = neutral)
        # x0_lambdas: blends initial embedding back in at each layer (init 0.0 = disabled)
        # Separate parameters so they can have different optimizer treatment
        if use_resid_lambdas:
            self.resid_lambdas = nn.Parameter(torch.ones(self.num_layers))  # fake init, real init in init_weights()
        else:
            self.register_buffer("resid_lambdas", torch.ones(self.num_layers), persistent=False)
        if use_x0_lambdas:
            self.x0_lambdas = nn.Parameter(torch.zeros(self.num_layers))  # fake init, real init in init_weights()
        else:
            self.register_buffer("x0_lambdas", torch.zeros(self.num_layers), persistent=False)

        self.rotary_seq_len = self.seq_len
        cos, sin = RoPE.precompute(self.rotary_seq_len, config.head_dim)
        self.register_buffer("cos", cos, persistent=False)  # persistent=False means it's not saved to the checkpoint
        self.register_buffer("sin", sin, persistent=False)

    @torch.no_grad()
    def init_weights(self):
        """
        Initialize the full model in this one function for maximum clarity.

        wte (embedding):     normal, std=1.0
        lm_head:             normal, std=0.001
        for each block:
            attn.c_q:        uniform, std=1/sqrt(n_embd)
            attn.c_k:        uniform, std=1/sqrt(n_embd)
            attn.c_v:        uniform, std=1/sqrt(n_embd)
            attn.c_proj:     zeros
            mlp.c_fc:        uniform, std=1/sqrt(n_embd)
            mlp.c_proj:      zeros
        """

        # Embedding and unembedding
        torch.nn.init.normal_(self.token_emb.weight, mean=0.0, std=1.0)
        torch.nn.init.normal_(self.lm_head.weight, mean=0.0, std=0.001)

        # Transformer blocks: uniform init with bound = sqrt(3) * std (same standard deviation as normal)
        n_embd = self.hidden_dim
        s = 3**0.5 * n_embd**-0.5  # sqrt(3) multiplier makes sure Uniform achieves the same std as Normal
        # pyrefly: ignore
        for block in self.blocks:
            block: Block
            torch.nn.init.uniform_(block.attn.c_q.weight, -s, s)  # weights use Uniform to avoid outliers
            torch.nn.init.uniform_(block.attn.c_k.weight, -s, s)
            torch.nn.init.uniform_(block.attn.c_v.weight, -s, s)
            torch.nn.init.zeros_(block.attn.c_proj.weight)  # projections are zero
            torch.nn.init.uniform_(block.mlp.c_fc.weight, -s, s)
            torch.nn.init.zeros_(block.mlp.c_proj.weight)

        # Per-layer scalars
        self.resid_lambdas.fill_(1.0)  # 1.0 => typical residual connections at init
        self.x0_lambdas.fill_(0.0)  # 0.0 => skip connection to input is disabled at init

        # Rotary embeddings
        head_dim = self.hidden_dim // self.num_heads
        self.cos, self.sin = RoPE.precompute(self.rotary_seq_len, head_dim, device=self.cos.device)

        # Cast token embeddings to bf16: optimizer can tolerate it and it saves memory
        if self.token_emb.weight.device.type == "cuda":
            self.token_emb.to(dtype=torch.bfloat16)

    def param_groups(
        self,
    ) -> dict[str, list[torch.nn.Parameter]]:
        # Separate out all parameters into 5 groups (matrix, embedding, lm_head, resid_lambdas, x0_lambdas)
        matrix_params = list(self.blocks.parameters())
        embedding_params = list(self.token_emb.parameters())
        lm_head_params = list(self.lm_head.parameters())
        resid_params = [self.resid_lambdas] if self.resid_lambdas.requires_grad else []
        x0_params = [self.x0_lambdas] if self.x0_lambdas.requires_grad else []
        assert len(list(self.parameters())) == (
            len(matrix_params) + len(embedding_params) + len(lm_head_params) + len(resid_params) + len(x0_params)
        )

        return {
            "matrix": matrix_params,
            "embedding": embedding_params,
            "lm_head": lm_head_params,
            "resid_lambdas": resid_params,
            "x0_lambdas": x0_params,
        }

    def forward(
        self,
        input: torch.Tensor,
        *,
        metadata: VarlenMetaData | None = None,
    ) -> torch.Tensor:
        B, T = input.size()

        cos_sin = self.cos[:, :T], self.sin[:, :T]  # truncate cache to current sequence length

        # Forward the trunk of the Transformer
        x = self.token_emb(input)
        x = self.norm(x)
        x0 = x  # save initial normalized embedding for x0 residual
        for block, resid_lambda, x0_lambda in zip(self.blocks, self.resid_lambdas, self.x0_lambdas):
            x = resid_lambda * x + x0_lambda * x0
            x = block(x, cos_sin, metadata=metadata)
        x = self.norm(x)
        logits = self.lm_head(x)  # (B, T, padded_vocab_size) <- very big tensor, large amount of memory
        logits = logits[..., : self.vocab_size]  # slice to remove padding
        if self.use_softcap:
            logits = self.softcap_logits(logits, softcap=15)

        return logits

    def softcap_logits(
        self,
        logits: torch.Tensor,
        softcap: float = 15,
    ) -> torch.Tensor:
        # Softcap the logits to the range [-softcap, softcap]
        logits = logits.float()  # switch to fp32 for logit softcap
        logits = softcap * torch.tanh(logits / softcap)  # squash the logits
        return logits



from collections.abc import Iterable
from functools import partial
from typing import Any, Literal, cast

import torch
from datasets import Dataset as HFDataset
from datasets import IterableDataset as HFIterableDataset
from datasets import disable_progress_bars
from datasets.distributed import split_dataset_by_node
from loguru import logger
from torch.utils.data import IterableDataset

from sarasa.data.tokenizer import BaseTokenizerWrapper
from sarasa.data.utils import prepare_varlen_metadata
from sarasa.utils import IGNORE_INDEX, rank, world_size


class HFTextDataset(IterableDataset):
    """
    A wrapper around HF datasets.

    Supports three strategies for creating input-label pairs:
    1. "streaming": Concatenate tokens from the dataset and yield input-label pairs of length seq_len + 1.
    2. "streaming_pad": Similar to "streaming", but if the buffer has fewer than seq_len + 1 tokens, pad with bos_token_id until it does.
    3. "document_pack": Create a buffer of tokens and repeatedly pick the longest sequences that fit within seq_len + 1, cropping if necessary.
    4. "document_pack_pad": Similar to "document_pack", but if no sequence can fit, pad with bos_token_id instead of cropping.
    """

    def __init__(
        self,
        dataset: HFDataset | HFIterableDataset,
        tokenizer: BaseTokenizerWrapper,
        seq_len: int,
        use_varlen: bool,
        strategy: Literal["streaming", "streaming_pad", "document_pack", "document_pack_pad"],
        infinite: bool = True,
        buffer_size: int = 1_000,
    ):
        if rank() != 0:
            disable_progress_bars()

        self.data = split_dataset_by_node(dataset, rank=rank(), world_size=world_size())
        self.tokenizer = tokenizer
        self.seq_len = seq_len
        self.infinite = infinite
        self.strategy = strategy
        self.buffer_size = buffer_size
        self.post_process_fn = (
            partial(prepare_varlen_metadata, bos_token_id=tokenizer.bos_token_id) if use_varlen else lambda x: x
        )

    def _text_processor(
        self,
        sample: dict,
    ) -> str:
        # Default text processor: extract 'text' field
        return sample["text"]

    def _streaming_iter(
        self,
        data_iter: Iterable[dict[str, Any]],
    ):
        buffer: list[int] = []

        for sample in data_iter:
            sample_text = self._text_processor(sample)
            sample_tokens = self.tokenizer.encode(sample_text)
            buffer.extend(sample_tokens)

            while len(buffer) >= 1 + self.seq_len:
                output = torch.tensor(buffer[: 1 + self.seq_len], dtype=torch.int64)
                buffer = buffer[1 + self.seq_len :]
                input = output[:-1]
                label = output[1:]
                yield self.post_process_fn({"input": input}), label

    def _streaming_pad_iter(
        self,
        data_iter: Iterable[dict[str, Any]],
    ):
        for sample in data_iter:
            sample_text = self._text_processor(sample)
            sample_tokens = self.tokenizer.encode(sample_text)

            pad_size = 0
            if len(sample_tokens) > self.seq_len + 1:
                sample_tokens = sample_tokens[: self.seq_len + 1]
            else:
                pad_size = self.seq_len + 1 - len(sample_tokens)
                sample_tokens += [self.tokenizer.bos_token_id] * pad_size

            input = sample_tokens[:-1]
            label = sample_tokens[1 : self.seq_len + 1 - pad_size] + [IGNORE_INDEX] * pad_size
            input = torch.tensor(input, dtype=torch.int64)
            label = torch.tensor(label, dtype=torch.int64)
            yield self.post_process_fn({"input": input}), label

    def _document_pack_iter(
        self,
        data_iter: Iterable[dict[str, Any]],
    ):
        # repeat picking the longest sequences from the buffer,  that fit local buffer (max length seq_len + 1)
        # if there's no sequence that can fit, pick the shortest one and crop it to fit / pad bos token to fit
        # this strategy is used in nanochat

        document_buffer: list[list[int]] = []
        output: list[int] = []
        pad_size = 0
        data_iter = iter(data_iter)

        while True:
            # fill the buffer
            while (sample := next(data_iter, None)) is not None and (len(document_buffer) < self.buffer_size):
                sample_text = self._text_processor(sample)
                sample_tokens = self.tokenizer.encode(sample_text)
                document_buffer.append(sample_tokens)

            if sample is None and len(document_buffer) == 0:
                break

            document_buffer.sort(key=len, reverse=True)

            i = 0
            while i < len(document_buffer):
                seq = document_buffer[i]
                if len(output) + len(seq) <= self.seq_len + 1:
                    output.extend(seq)
                    document_buffer.pop(i)
                else:
                    i += 1

            # if no sequence can fit, pick the shortest one and crop it to fit / pad bos token to fit
            if len(output) < self.seq_len + 1:
                if self.strategy == "document_pack":
                    seq = document_buffer[-1]
                    output.extend(seq)
                    output = output[: self.seq_len + 1]
                    document_buffer.pop(-1)
                else:
                    pad_size = self.seq_len + 1 - len(output)
                    output += [self.tokenizer.bos_token_id] * pad_size

            output = output[: self.seq_len + 1]
            input = output[:-1]
            label = output[1 : self.seq_len + 1 - pad_size] + [IGNORE_INDEX] * pad_size
            input = torch.tensor(input, dtype=torch.int64)
            label = torch.tensor(label, dtype=torch.int64)
            yield self.post_process_fn({"input": input}), label

            output = []
            pad_size = 0

    def __iter__(self):
        while True:
            data_iter = cast(Iterable, iter(self.data))
            match self.strategy:
                case "streaming":
                    yield from self._streaming_iter(data_iter)
                case "streaming_pad":
                    yield from self._streaming_pad_iter(data_iter)
                case "document_pack" | "document_pack_pad":
                    yield from self._document_pack_iter(data_iter)
                case _:
                    raise ValueError(f"Invalid packing strategy: {self.strategy}")

            if not self.infinite:
                break
            else:
                # Reset offset for the next iteration
                logger.warning("Dataset is being re-looped")
                if hasattr(self.data, "set_epoch") and hasattr(self.data, "epoch"):
                    self.data.set_epoch(self.data.epoch + 1)



from __future__ import annotations

import inspect
from functools import partial
from typing import Callable, ClassVar, NamedTuple, cast

import torch
from torch import nn
from torch.nn import functional as F
from torch.nn.attention.varlen import varlen_attn as _varlen_attn
from torch.types import Number

from sarasa.models import ModelConfig
from sarasa.models.utils import RoPE

if inspect.signature(_varlen_attn).parameters.get("window_size") is not None:
    # torch>2.10
    # after the release of torch 2.11, this branch can be removed
    varlen_attn = partial(_varlen_attn, window_size=(-1, 0))
else:
    # torch==2.10
    varlen_attn = partial(_varlen_attn, is_causal=True)


class VarlenMetaData(NamedTuple):
    cu_seq_q: torch.Tensor
    cu_seq_k: torch.Tensor
    max_q: Number
    max_k: Number

    def to(
        self,
        device: torch.device,
        non_blocking: bool = False,
    ) -> VarlenMetaData:
        return self._replace(
            cu_seq_q=self.cu_seq_q.to(device, non_blocking=non_blocking),
            cu_seq_k=self.cu_seq_k.to(device, non_blocking=non_blocking),
        )


class VarlenAttention(nn.Module):
    compiled_varlen: ClassVar[Callable] = torch.compile(
        varlen_attn,
        mode="max-autotune-no-cudagraphs",
    )

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        *,
        metadata: VarlenMetaData,
    ) -> torch.Tensor:

        out = VarlenAttention.compiled_varlen(
            query.transpose(1, 2).flatten(0, 1),  # (B*T, num_heads, head_dim)
            key.transpose(1, 2).flatten(0, 1),
            value.transpose(1, 2).flatten(0, 1),
            cu_seq_q=metadata.cu_seq_q.view(-1),
            cu_seq_k=metadata.cu_seq_k.view(-1),
            max_q=int(metadata.max_q),
            max_k=int(metadata.max_k),
        )  # (B*T, num_heads, head_dim)
        return out.reshape(query.size(0), query.size(2), -1)  # (B, T, num_heads * head_dim)


class SDPAttention(nn.Module):
    def __init__(
        self,
        is_causal: bool,
        enable_gqa: bool,
    ):
        super().__init__()
        self.is_causal = is_causal
        self.enable_gqa = enable_gqa

        if nn.attention.current_flash_attention_impl() == "FA4":
            self.sdpa_backends = nn.attention.SDPBackend.FLASH_ATTENTION
        else:
            # type: ignore
            self.sdpa_backends = [
                nn.attention.SDPBackend.CUDNN_ATTENTION,
                nn.attention.SDPBackend.FLASH_ATTENTION,
                nn.attention.SDPBackend.MATH,
            ]

    def forward(
        self,
        query: torch.Tensor,
        key: torch.Tensor,
        value: torch.Tensor,
        *,
        metadata=None,
    ) -> torch.Tensor:
        assert metadata is None
        with nn.attention.sdpa_kernel(self.sdpa_backends):
            out = F.scaled_dot_product_attention(
                query,  # (B, num_heads, T, head_dim)
                key,
                value,
                is_causal=self.is_causal,
                enable_gqa=self.enable_gqa,
            )
        return out.transpose(1, 2).reshape(query.size(0), query.size(2), -1)  # (B, T, num_heads * head_dim)


class CausalSelfAttention(nn.Module):
    def __init__(
        self,
        config: ModelConfig,
        layer_idx: int | None = None,
    ):
        super().__init__()
        self.layer_idx = cast(int, layer_idx)
        self.num_heads = cast(int, config.num_heads)
        self.num_kv_heads = cast(int, config.num_kv_heads)
        self.hidden_dim = cast(int, config.hidden_dim)
        self.head_dim = self.hidden_dim // self.num_heads
        self.c_q = nn.Linear(self.hidden_dim, self.num_heads * self.head_dim, bias=False)
        self.c_k = nn.Linear(self.hidden_dim, self.num_kv_heads * self.head_dim, bias=False)
        self.c_v = nn.Linear(self.hidden_dim, self.num_kv_heads * self.head_dim, bias=False)
        self.c_proj = nn.Linear(self.hidden_dim, self.hidden_dim, bias=False)
        self.qk_norm = (
            nn.RMSNorm(self.head_dim, eps=config.rms_eps, elementwise_affine=config.rms_learnable)
            if config.qk_norm
            else nn.Identity()
        )

        self.attn: nn.Module
        if config.attn_type == "varlen":
            self.attn = VarlenAttention()
        else:
            self.attn = SDPAttention(is_causal=True, enable_gqa=self.num_heads != self.num_kv_heads)

    def forward(
        self,
        x: torch.Tensor,
        cos_sin: tuple[torch.Tensor, torch.Tensor],
        *,
        metadata: VarlenMetaData | None = None,
    ) -> torch.Tensor:
        B, T, C = x.size()

        # Project the input to get queries, keys, and values
        q = self.c_q(x).view(B, T, self.num_heads, self.head_dim)
        k = self.c_k(x).view(B, T, self.num_kv_heads, self.head_dim)
        v = self.c_v(x).view(B, T, self.num_kv_heads, self.head_dim)

        # Apply Rotary Embeddings to queries and keys to get relative positional encoding
        cos, sin = cos_sin
        q, k = RoPE.apply(q, cos, sin), RoPE.apply(k, cos, sin)
        q, k = self.qk_norm(q), self.qk_norm(k)
        y = self.attn(
            q.transpose(1, 2),
            k.transpose(1, 2),
            v.transpose(1, 2),
            metadata=metadata,
        )  # (B, T, num_heads * head_dim)
        y = self.c_proj(y)
        return y



import typing

import pytest
import torch

from sarasa.models import ModelConfig


@pytest.mark.parametrize("name", typing.get_type_hints(ModelConfig)["name"].__args__)
@torch.no_grad()
def test_model_shape_sdpa(name):
    config = ModelConfig(
        name=name,
        num_layers=4,
        head_dim=64,
        vocab_size=32,
        seq_len=16,
        attn_type="sdpa",
    )
    with torch.device("meta"):
        model = config.create()
        input = torch.randint(0, config.vocab_size, (1, config.seq_len), dtype=torch.long)
        output = model(input)
    assert output.shape == (1, config.seq_len, config.vocab_size)


@pytest.mark.parametrize("name", typing.get_type_hints(ModelConfig)["name"].__args__)
@torch.no_grad()
def test_model_shape_varlen(name):
    from sarasa.data.utils import prepare_varlen_metadata

    config = ModelConfig(
        name=name,
        num_layers=4,
        head_dim=64,
        vocab_size=32,
        seq_len=16,
        attn_type="varlen",
    )
    input = torch.tensor([0, 1, 1, 1, 0, 1, 1], dtype=torch.long)
    input_dict = {"input": input}
    input_dict = prepare_varlen_metadata(input_dict, bos_token_id=0)
    input_dict["input"] = input_dict["input"].unsqueeze(0)  # collate
    with torch.device("meta"):
        model = config.create()
        output = model(**input_dict)
    assert output.shape == (1, input.size(0), config.vocab_size)



from .config import AdamW as AdamW
from .config import Checkpoint as Checkpoint
from .config import Config as Config
from .config import Data as Data
from .config import Distributed as Distributed
from .config import LRScheduler as LRScheduler
from .config import Metrics as Metrics
from .config import Model as Model
from .config import Train as Train
from .train import Trainer as Trainer



import json
from pathlib import Path

from tokenizers import Tokenizer

SPECIAL_TOKENS = [
    # special tokens excluding bos
    "<|user_start|>",
    "<|user_end|>",
    "<|assistant_start|>",
    "<|assistant_end|>",
]


class BaseTokenizerWrapper:
    bos_token_id: int

    def encode(self, text: str, **kwargs) -> list[int]:
        raise NotImplementedError

    def decode(self, token_ids: list[int], **kwargs) -> str:
        raise NotImplementedError

    def __len__(self) -> int:
        raise NotImplementedError


class HFTokenizerWrapper(BaseTokenizerWrapper):
    def __init__(
        self,
        tokenizer_path: Path,
    ):
        self.tokenizer = Tokenizer.from_file(str(tokenizer_path / "tokenizer.json"))
        with (tokenizer_path / "tokenizer_config.json").open("r") as f:
            config = json.load(f)

        bos_token = self._get_tokens_from_config(config.get("bos_token", None))
        if bos_token is None:
            raise ValueError("BOS token must be specified in the tokenizer config.")

        # check if tokenizer adds bos token automatically
        test_encoding = self.tokenizer.encode("test").ids
        self.bos_token_id = self.tokenizer.token_to_id(bos_token)
        self.need_bos = self.bos_token_id not in test_encoding

        self.tokenizer.add_special_tokens(SPECIAL_TOKENS)

    def _get_tokens_from_config(
        self,
        token: dict[str, str] | str | None,
    ) -> str | None:
        if isinstance(token, dict):
            token = token["content"]
        return token

    def encode(
        self,
        text: str,
        **kwargs,
    ) -> list[int]:
        token_ids = self.tokenizer.encode(text, **kwargs).ids

        if self.need_bos:
            token_ids = [self.bos_token_id] + token_ids

        return token_ids

    def decode(
        self,
        token_ids: list[int],
        **kwargs,
    ) -> str:
        return self.tokenizer.decode(token_ids, **kwargs)

    def __len__(self) -> int:
        return self.tokenizer.get_vocab_size()

    def render_messages(
        self,
        messages: list[dict[str, str]],
        max_length: int,
    ) -> tuple[list[int], list[int]]:
        """Render messages into ids and mask.
        We assume each `messages` has the following structure:

        [
            {"role": "user" | "assistant", "content": "..."},
            ...,
        ]

        """

        ids, mask = [self.bos_token_id], [0]

        for i, message in enumerate(messages):
            role = message["role"]
            content = message["content"]

            assert role in ["user", "assistant"], f"Unknown role: {role}"
            assert isinstance(content, str), "Message content should be a string."

            if role == "user":
                _ids = self.encode(f"<|user_start|>{content}<|user_end|>")[1:]  # remove bos
                ids.extend(_ids)
                mask.extend([0] * len(_ids))

            else:  # assistant
                _ids = self.encode(f"<|assistant_start|>{content}<|assistant_end|>")[1:]  # remove bos
                ids.extend(_ids)
                mask.extend([1] * len(_ids))

        # Truncate from the left if exceeds max_length
        ids = ids[:max_length]
        mask = mask[:max_length]

        return ids, mask



import torch


class RoPE:
    @staticmethod
    def precompute(
        seq_len: int,
        head_dim: int,
        device: torch.device | None = None,
        base: float = 10000,
    ):
        channel_range = torch.arange(0, head_dim, 2, dtype=torch.float32, device=device)
        inv_freq = 1.0 / (base ** (channel_range / head_dim))
        t = torch.arange(seq_len, dtype=torch.float32, device=device)
        freqs = torch.outer(t, inv_freq)[None, :, None, :]
        cos, sin = freqs.cos(), freqs.sin()
        cos, sin = cos.bfloat16(), sin.bfloat16()
        return cos, sin

    @staticmethod
    def apply(
        x: torch.Tensor,
        cos: torch.Tensor,
        sin: torch.Tensor,
    ) -> torch.Tensor:
        assert x.ndim == 4
        x1, x2 = x.chunk(2, dim=-1)
        y1 = x1 * cos + x2 * sin
        y2 = x1 * (-sin) + x2 * cos
        return torch.cat([y1, y2], 3)



import dataclasses
import subprocess
import time
from pathlib import Path
from typing import Any

import torch
from loguru import logger
from torch._utils import _get_device_module

from sarasa.config import Config
from sarasa.utils import rank


# ported from torchtitan
# hardcoded BF16 type peak flops for NVIDIA A100, H100, H200, B200 GPU and AMD MI250, MI300X, MI325X, MI355X and Intel PVC
def get_peak_flops(device_name: str) -> float:
    try:
        # Run the lspci command and capture the output
        result = subprocess.run(["lspci"], stdout=subprocess.PIPE, text=True)
        # Filter the output for lines containing both "NVIDIA" and "H100"
        filtered_lines = [line for line in result.stdout.splitlines() if "NVIDIA" in line and "H100" in line]
        # Join all filtered lines into a single string
        device_name = " ".join(filtered_lines) or device_name
    except FileNotFoundError as e:
        logger.warning(f"Error running lspci: {e}, fallback to use device_name")
    if "A100" in device_name:
        # data from https://www.nvidia.com/en-us/data-center/a100/
        return 312e12
    elif "H100" in device_name:
        # data from https://www.nvidia.com/en-us/data-center/h100/
        # NOTE: Specifications are one-half lower without sparsity.
        if "NVL" in device_name:
            return 835e12
        elif "PCIe" in device_name:
            return 756e12
        else:  # for H100 SXM and other variants
            return 989e12
    elif "H200" in device_name:
        # data from https://www.nvidia.com/en-us/data-center/h200/
        return 989e12
    elif "B200" in device_name:
        # data from https://nvdam.widen.net/s/wwnsxrhm2w/blackwell-datasheet-3384703
        return 2.25e15
    elif "MI355X" in device_name:
        # MI355X data from https://www.amd.com/en/products/accelerators/instinct/mi350/mi355x.html
        return 2500e12
    elif "MI300X" in device_name or "MI325X" in device_name:
        # MI300X data from https://www.amd.com/en/products/accelerators/instinct/mi300/mi300x.html
        # MI325X data from https://www.amd.com/en/products/accelerators/instinct/mi300/mi325x.html
        return 1300e12
    elif "MI250X" in device_name:
        # data from https://www.amd.com/en/products/accelerators/instinct/mi200/mi250x.html (per GCD)
        return 191.5e12
    elif "Data Center GPU Max 1550" in device_name:
        # Also known as Ponte Vecchio (PVC).
        # data from https://www.intel.com/content/www/us/en/docs/oneapi/optimization-guide-gpu/2025-0/intel-xe-gpu-architecture.html
        # Dot Product Accumulate Systolic (DPAS):
        # - Freq: 1300MHz
        # - #ops: 512
        # Full EU mode (i.e. 512 max compute units): 340.8 TFLOPS (BF16)
        # Standard EU mode (i.e. 448 max compute units): 298.2 TFLOPS (BF16)
        max_comp_units = torch.xpu.get_device_properties("xpu").max_compute_units
        return 512 * max_comp_units * 1300 * 10**6
    elif "l40s" in device_name:
        # data from: "https://resources.nvidia.com/en-us-l40s/l40s-datasheet-28413"
        return 362e12

    else:  # for other GPU types, assume A100
        logger.warning(f"Peak flops undefined for: {device_name}, fallback to A100")
        return 312e12


@dataclasses.dataclass(slots=True)
class DevMemStats:
    max_active_gib: float
    max_active_perc: float
    max_reserved_gib: float
    max_reserved_perc: float
    num_alloc_retries: int
    num_ooms: int


class DeviceMemoryMonitor:
    def __init__(
        self,
        device: torch.device,
    ) -> None:
        self.device = device
        try:
            _, self.total_mem = torch.accelerator.get_memory_info(self.device)
            self.device_name = _get_device_module(self.device.type).get_device_name(self.device)
        except RuntimeError:
            if self.device.type == "mps":
                self.total_mem = torch.mps.recommended_max_memory()
                self.device_name = "Apple Silicon GPU"
            else:
                raise NotImplementedError(f"Device memory monitor not implemented for device type: {self.device.type}")

        self.reset_peak_stats()
        try:
            torch.accelerator.empty_cache()
        except RuntimeError:
            logger.error(f"Failed to empty cache for device type: {self.device.type}")

    @staticmethod
    def to_gib(bytes: int) -> float:
        return bytes / (1024**3)

    def reset_peak_stats(self) -> None:
        try:
            torch.accelerator.reset_peak_memory_stats(self.device)
        except RuntimeError:
            logger.error(f"Failed to reset peak memory stats for device type: {self.device.type}")

    def get_peak_stats(self) -> DevMemStats:
        try:
            info = torch.accelerator.memory_stats(self.device)
        except RuntimeError:
            logger.error(f"Failed to get peak memory stats for device type: {self.device.type}")
            info = {}

        max_active = info.get("active_bytes.all.peak", -1)
        max_reserved = info.get("reserved_bytes.all.peak", -1)
        num_retries = info.get("num_alloc_retries", -1)
        num_ooms = info.get("num_ooms", -1)

        if num_retries > 0:
            logger.warning(f"{num_retries} {self.device.type.upper()} memory allocation retries.")
        if num_ooms > 0:
            logger.warning(f"{num_ooms} {self.device.type.upper()} OOM errors thrown.")

        return DevMemStats(
            max_active_gib=self.to_gib(max_active),
            max_active_perc=max_active / self.total_mem * 100,
            max_reserved_gib=self.to_gib(max_reserved),
            max_reserved_perc=max_reserved / self.total_mem * 100,
            num_alloc_retries=num_retries,
            num_ooms=num_ooms,
        )


class BaseReporter:
    def config(
        self,
        config: dict[str, Any],
    ) -> None:
        raise NotImplementedError()

    def log(
        self,
        metrics: dict[str, Any],
        step: int,
    ) -> None:
        raise NotImplementedError()

    def close(self) -> None:
        raise NotImplementedError()


class TensorboardReporter(BaseReporter):
    def __init__(
        self,
        log_dir: Path,
    ) -> None:
        from torch.utils.tensorboard import SummaryWriter

        self.writer = SummaryWriter(log_dir=log_dir, max_queue=1000)

        logger.info(f"TensorBoard log is available at {log_dir}")

    def config(
        self,
        config: dict[str, Any],
    ) -> None:
        for k, v in config.items():
            self.writer.add_text(f"config/{k}", str(v))

    def log(
        self,
        metrics: dict[str, float],
        step: int,
    ) -> None:
        for k, v in metrics.items():
            self.writer.add_scalar(k, v, step)

    def close(self) -> None:
        self.writer.close()


class MetricsProcessor:
    def __init__(
        self,
        config: Config,
        device: torch.device,
        flops_per_token: int,
    ) -> None:
        self.reporters = []
        if config.metrics.all_node or rank() == 0:
            if config.metrics.use_tensorboard:
                log_dir = config.output_dir / "tensorboard" if config.output_dir else Path("./tensorboard")
                self.reporters.append(TensorboardReporter(log_dir=log_dir))

        for reporter in self.reporters:
            reporter.config(config=dataclasses.asdict(config))

        self.device_mem_monitor = DeviceMemoryMonitor(device)
        self.freq = config.metrics.freq
        self.time_last_log = time.perf_counter()
        gpu_peak_flops = get_peak_flops(self.device_mem_monitor.device_name)
        logger.info(f"Detected device: {self.device_mem_monitor.device_name}, Peak FLOPS: {gpu_peak_flops}")
        self.gpu_peak_flops = gpu_peak_flops
        self.ntokens_since_last_log = 0
        self.flops_per_token = flops_per_token
        self.data_load_times: list[float] = []
        self.reset()

    def trigger(
        self,
        step: int,
    ) -> bool:
        return step == 1 or step % self.freq == 0

    def log(
        self,
        step: int,
        global_avg_loss: float,
        global_max_loss: float,
        extra_metrics: dict[str, float] | None = None,
    ) -> None:
        time_delta = time.perf_counter() - self.time_last_log
        device_mem_stats = self.device_mem_monitor.get_peak_stats()
        time_ete = time_delta / self.freq
        time_data_load = sum(self.data_load_times) / len(self.data_load_times) if self.data_load_times else 0.0
        time_data_load_perc = 100 * time_data_load / time_ete if time_ete > 0 else 0.0

        metrics = {
            "loss/avg": global_avg_loss,
            "loss/max": global_max_loss,
            "memory/max_active(GiB)": device_mem_stats.max_active_gib,
            "memory/max_active(%)": device_mem_stats.max_active_perc,
            "memory/max_reserved(GiB)": device_mem_stats.max_reserved_gib,
            "memory/max_reserved(%)": device_mem_stats.max_reserved_perc,
            "memory/num_alloc_retries": device_mem_stats.num_alloc_retries,
            "memory/num_ooms": device_mem_stats.num_ooms,
            "time/end-to-end(s)": time_ete,
            "time/data_load(s)": time_data_load,
            "time/data_load(%)": time_data_load_perc,
        }

        log = (
            f"[Step {step:>10}] loss: {global_avg_loss:.4f}, memory: {device_mem_stats.max_reserved_gib:.2f} GiB, "
            f"time(s): {time_ete:.2f}sec (data load ratio: {time_data_load_perc:.1f}%)"
        )

        if extra_metrics is not None:
            metrics.update(extra_metrics)

        if self.flops_per_token > 0:
            tps = self.ntokens_since_last_log / time_delta
            mfu = 100 * self.flops_per_token * tps / self.gpu_peak_flops
            tflops = self.flops_per_token * tps / 1e12

            metrics.update({
                "throughput(tps)": tps,
                "tflops": tflops,
                "mfu(%)": mfu,
            })
            log += f", tflops: {tflops:.2f}, mfu: {mfu:.2f}%"

        for reporter in self.reporters:
            reporter.log(metrics, step)

        logger.info(log)

        self.reset()

    def val_log(
        self,
        step: int,
        val_loss: float,
        extra_metrics: dict[str, float] | None = None,
    ) -> None:
        time_delta = time.perf_counter() - self.time_last_log
        device_mem_stats = self.device_mem_monitor.get_peak_stats()

        metrics = {
            "val/loss": val_loss,
            "val/memory/max_active(GiB)": device_mem_stats.max_active_gib,
            "val/memory/max_active(%)": device_mem_stats.max_active_perc,
            "val/memory/max_reserved(GiB)": device_mem_stats.max_reserved_gib,
            "val/memory/max_reserved(%)": device_mem_stats.max_reserved_perc,
            "val/time(s)": time_delta,
        }

        if extra_metrics is not None:
            metrics.update(extra_metrics)

        for reporter in self.reporters:
            reporter.log(metrics, step)

        logger.info(f"[Eval Step {step:>10}] val loss: {val_loss:.4f}, val time: {time_delta:.2f} sec")

        self.reset()

    def reset(self) -> None:
        self.ntokens_since_last_log = 0
        self.data_load_times.clear()
        self.time_last_log = time.perf_counter()
        self.device_mem_monitor.reset_peak_stats()

    def close(self) -> None:
        for reporter in self.reporters:
            reporter.close()



from typing import Any

import torch

from sarasa.models.attention import VarlenMetaData


def prepare_varlen_metadata(
    input_dict: dict[str, Any],
    bos_token_id: int,
) -> dict[str, torch.Tensor | VarlenMetaData]:
    # add metadata for varlen attention

    assert bos_token_id is not None

    input = input_dict["input"]  # T

    (bos_positions,) = (input == bos_token_id).nonzero(as_tuple=True)
    bos_positions = bos_positions.to(torch.int32)
    if len(bos_positions) == 0 or bos_positions[0] != 0:
        bos_positions = torch.cat((bos_positions.new_tensor([0]), bos_positions))
    cu_seq = torch.cat((
        bos_positions,
        bos_positions.new_tensor([input.size(0)]),
    ))
    max_seqlen = torch.diff(cu_seq).max().item()
    metadata = VarlenMetaData(
        cu_seq_q=cu_seq,
        cu_seq_k=cu_seq,
        max_q=max_seqlen,
        max_k=max_seqlen,
    )
    input_dict["metadata"] = metadata
    return input_dict



import contextlib
import os
import time
from typing import Iterator

import torch
import torch.distributed as dist
from loguru import logger
from torch.distributed.elastic.multiprocessing.errors import record
from torch.nn import functional as F

from sarasa.activation_checkpoint import apply_op_sac
from sarasa.checkpoint import Checkpointer
from sarasa.config import Config
from sarasa.evaluate import Evaluator
from sarasa.metrics import MetricsProcessor
from sarasa.utils import (
    IGNORE_INDEX,
    GarbageCollector,
    apply_distributed,
    init_distributed,
    set_dtype,
    setup_logger,
    setup_profiler,
    update_timeout,
    world_size,
)


class Trainer:
    @record
    def __init__(
        self,
        config: Config,
    ) -> None:
        self.config = config

        # only rank 0 logs
        setup_logger(config)

        logger.info(f"Initializing Trainer with config: {self.config}")

        # set seed
        torch.manual_seed(config.seed)
        os.environ["PYTHONHASHSEED"] = str(config.seed % 2**32)

        # setup device
        torch.accelerator.set_device_index(int(os.environ.get("LOCAL_RANK", 0)))
        self.device = torch.accelerator.current_accelerator(check_available=True) or torch.device("cpu")

        self.gc = GarbageCollector(config.train.gc_freq)

        # setup distributed
        init_distributed(config.distributed.backend, config.distributed.init_timeout_seconds)

        # setup data and tokenizer -> use vocab size for model setup
        data = config.data.create(
            batch_size=config.train.local_batch_size,
            val_cfg=config.evaluate,
            use_varlen=config.model.attn_type == "varlen",
        )
        self.data_loader = data["train_loader"]  # setup data loader
        val_loader = data.get("val_loader", None)

        self.config.model.vocab_size = len(data["tokenizer"])

        # setup model, optimizer, lr scheduler
        with torch.device("meta"), set_dtype(getattr(torch, config.train.dtype)):
            self.model = self.config.model.create()
            num_params, flops_per_token = self.model.num_params_flops
            model_size = num_params / 1e9
            model_size, unit = (num_params / 1e6, "M") if model_size < 1 else (model_size, "B")
            logger.info(f"Model created with {model_size:.2f}{unit} parameters")

        # following torchtitan, (S)AC -> compilation -> distributed wrapping
        if config.train.use_sac:
            logger.info("Applying Selective Activation Checkpointing (SAC)")
            for i, block in enumerate(self.model.blocks):
                self.model.blocks[i] = apply_op_sac(block)

        if config.train.compile:
            logger.info("Compiling the model")
            for block in self.model.blocks:
                block.compile(fullgraph=True)
            self.model.compile()
            self.loss_fn = torch.compile(self.loss_fn, fullgraph=True, dynamic=False)

        if world_size() > 1:
            apply_distributed(
                config.distributed,
                self.model,
                device=self.device,
                param_dtype=getattr(torch, config.train.amp_dtype),
                reduce_dtype=getattr(torch, config.train.dtype),
                compile=config.train.compile,
            )

        self.model.to_empty(device=self.device)
        self.model.init_weights()

        self.optimizer = self.config.optim.create(self.model)
        self.lr_scheduler = self.config.lr_scheduler.create(self.optimizer, config.train.steps)

        self.step = 0
        self.grad_accum_steps = config.train.global_batch_size // (config.train.local_batch_size * world_size())
        if self.grad_accum_steps > 1:
            logger.info(f"Gradient accumulation step is set to: {self.grad_accum_steps}")

        self.amp_context: contextlib.AbstractContextManager = contextlib.nullcontext()
        if world_size() == 1 or config.distributed.dp_shard_degree != -1:
            self.amp_context = torch.autocast(
                device_type=self.device.type,
                dtype=getattr(torch, config.train.amp_dtype),
            )

        self.profile_context = setup_profiler(self.config.profile, self.device, save_dir=self.config.output_dir)

        # setup metrics, checkpointer, evaluator
        self.metrics_processor = MetricsProcessor(config, self.device, flops_per_token)
        self.checkpointer = Checkpointer(config, self.model) if config.checkpoint.freq > 0 else None
        self.evaluator = (
            Evaluator(config.evaluate, val_loader, self.amp_context, self.metrics_processor, self.loss_fn, self.device)
            if val_loader is not None
            else None
        )

        dev_mem_stats = self.metrics_processor.device_mem_monitor.get_peak_stats()
        logger.info(
            f"{self.device.type.upper()} memory: {dev_mem_stats.max_reserved_gib:.2f} GiB for model initialization"
        )

        if config.train.use_fa4:
            logger.info("Using FA4 flash attention")
            try:
                torch.nn.attention.activate_flash_attention_impl("FA4")
            except Exception as e:
                logger.warning(
                    f"Failed to activate FA4 flash attention: {e}. Install sarasa with `flash_attn` extra for better performance."
                )

    @record
    def train(self):
        try:
            logger.info("Starting training...")

            self.model.train()
            with self.profile_context as profiler:
                data_iter = self.batch_generator(self.data_loader)
                for _ in range(self.config.train.steps):
                    self.step += 1
                    self.gc.collect(self.step)
                    try:
                        self.train_step(data_iter)
                    except StopIteration:
                        logger.warning("Data loader exhausted during training.")
                        break

                    if self.checkpointer is not None and self.checkpointer.trigger(self.step):
                        self.checkpointer.save(self.step)

                    if self.evaluator is not None and self.evaluator.trigger(self.step):
                        logger.info("Starting evaluation...")
                        self.evaluator.evaluate(self.model, self.step)

                    profiler.step()

                    if world_size() > 1 and self.step == 1:
                        update_timeout(self.config.distributed.train_timeout_seconds, self.device)

            logger.info("Training completed.")
        finally:
            logger.info("Cleaning up trainer...")
            self.close()

    def batch_generator(
        self,
        data_iter: Iterator[tuple[dict[str, torch.Tensor], torch.Tensor]],
    ) -> Iterator[tuple[dict[str, torch.Tensor], torch.Tensor]]:
        data_iter = iter(data_iter)
        while True:
            begin = time.perf_counter()
            batch = next(data_iter)
            input_dict, target = batch
            self.metrics_processor.ntokens_since_last_log += target.numel()
            self.metrics_processor.data_load_times.append(time.perf_counter() - begin)
            yield input_dict, target

    def train_step(
        self,
        batch_iter: Iterator[tuple[dict[str, torch.Tensor], torch.Tensor]],
    ) -> None:
        self.optimizer.zero_grad()

        micro_batches = []
        valid_tokens = torch.tensor(0, dtype=torch.long)
        for _ in range(self.grad_accum_steps):
            input_dict, target = next(batch_iter)
            valid_tokens += (target != IGNORE_INDEX).sum()
            micro_batches.append((input_dict, target))

        valid_tokens = valid_tokens.to(self.device)
        if world_size() > 1:
            dist.all_reduce(valid_tokens, op=dist.ReduceOp.SUM)

        losses = []
        for input_dict, target in micro_batches:
            input_dict = {
                k: v.to(self.device, non_blocking=(self.device.type == "cuda")) for k, v in input_dict.items()
            }
            target = target.to(self.device, non_blocking=(self.device.type == "cuda"))

            with self.amp_context:
                pred = self.model(**input_dict)
                loss = self.loss_fn(pred, target) / valid_tokens

            del pred
            loss.backward()
            losses.append(loss.detach())

        if self.config.train.grad_clip is not None:
            torch.nn.utils.clip_grad_norm_(
                self.model.parameters(),
                self.config.train.grad_clip,
                foreach=self.device.type == "cuda",
            )

        if self.checkpointer is not None:
            self.checkpointer.wait_for_staging()

        self.optimizer.step()
        self.lr_scheduler.step()

        loss = torch.stack(losses).sum()

        if not self.metrics_processor.trigger(self.step):
            return

        if world_size() > 1:
            avg_loss = loss.clone()
            dist.all_reduce(avg_loss, op=dist.ReduceOp.SUM)
            max_loss = loss.clone()
            dist.all_reduce(max_loss, op=dist.ReduceOp.MAX)
        else:
            avg_loss = max_loss = loss

        with torch.no_grad():
            grad_norm = torch.nn.utils.get_total_norm(self.model.parameters(), foreach=self.device.type == "cuda")

        lr = self.lr_scheduler.get_last_lr()[0]
        self.metrics_processor.log(
            self.step,
            global_avg_loss=avg_loss.item(),
            global_max_loss=max_loss.item(),
            extra_metrics={
                "grad_norm": grad_norm.item() if grad_norm >= 0 else float("nan"),
                "lr": lr,
            },
        )

    def loss_fn(
        self,
        pred: torch.Tensor,
        target: torch.Tensor,
    ) -> torch.Tensor:
        return F.cross_entropy(
            pred.flatten(0, 1).float(),
            target.flatten(0, 1),
            ignore_index=IGNORE_INDEX,
            reduction="sum",
        )

    def close(self) -> None:
        if self.checkpointer is not None:
            self.checkpointer.close()

        if self.metrics_processor is not None:
            self.metrics_processor.close()

        # cleanup distributed
        if world_size() > 1:
            try:
                dist.destroy_process_group()
            except Exception as e:
                logger.warning(f"Failed to destroy process group: {e}")

    def __del__(self):
        self.close()
